import numpy as np


def check_equals(a1, a2, sort=False):
    """
    checks that two arrays are the same

    :param sort: if True, sorts the arrays first
    """

    assert len(a1) == len(a2)
    assert a1.shape == a2.shape

    if sort:
        a1.sort()
        a2.sort()

    assert (a1 == a2).sum() == a1.size


def check_same_sets(sa1, sa2):
    """
    checks that two arrays contain the same items, irregardless of sort & repeats
    """
    a1 = np.unique(sa1)
    a2 = np.unique(sa2)

    check_equals(a1, a2, sort=True)


def check_no_dups(sa):
    """
    checks that array does not contain duplicates
    """
    assert len(sa) == len(np.unique(sa))


def get_unique(df, cols):
    df = df[cols]
    df = df.drop_duplicates(cols)
    df = df.sort_values(cols)
    return df



from collections import Counter
import fnmatch

from caseconverter import snakecase
import pandas
import numpy as np
import pandas as pd
from slugify import slugify

from . import xsettings
from . import xagg


def x_gen_grouped_counter(df, group_on=None, sort_on=None, start=0):
    if sort_on:
        df = df.sort_values(sort_on)

    return df.groupby(group_on).cumcount() + start


def x_filter_by_keys(df, key_values):
    df2 = df.copy()
    for k, v in key_values.iteritems():
        if k in df2.columns:
            df2 = df2[df2[k] == v].copy()

    return df2


def x_iter_groups(df, on):
    if not on:
        yield df, [], ""
        return

    if isinstance(on, str):
        on = [on]

    df = df.sort_values(on)
    groups = df.groupby(on)
    for group_keys, df_group in groups:
        if not isinstance(group_keys, list) and not isinstance(group_keys, tuple):
            group_keys = [group_keys]

        group_title = ", ".join([f"{k}={v}" for k,v in zip(on, group_keys)])
        series_group_keys = pd.Series(data=group_keys, index=on)
        yield df_group, series_group_keys, group_title


def x_merge(df_self, right, x_drop_dup_cols=True, drop_diff_cols=False, log=None, **kwargs):
    """
    A wrapper around pd.merge().
    Gets rid of as many duplicate columns ('_x', '_y') as possible after the merge.
    """
    
    df = df_self.merge(right, **kwargs)

    if log:
        on = kwargs['on']
        kleft = df_self[on].drop_duplicates()
        kright = right[on].drop_duplicates()
        kleft['xleft'] = 1
        kright['xright'] = 1

        both = kleft.merge(kright, on=on, how='outer')
        both[['xleft', 'xright']] = both[['xleft', 'xright']].fillna(0)
        both = both.sort_values(on)
        left_only = both[(both.xleft == 1) & (both.xright == 0)]
        if len(left_only):
            left_only.to_csv(xsettings.OUTPUT_PATH.joinpath(f"{log}-left_only.csv"), index=False)
        right_only = both[(both.xleft == 0) & (both.xright == 1)]
        if len(right_only):
            right_only.to_csv(xsettings.OUTPUT_PATH.joinpath(f"{log}-right_only.csv"), index=False)

    if not x_drop_dup_cols:
        return df

    suffixes = kwargs.get('suffixes', ('_x', '_y'))
    s1, s2 = suffixes
    for col1 in sorted(df.columns):
        if col1.endswith(s1):
            col_new = col1[:-1*len(s1)]
            col2 = col_new + s2

            if col_new in df.columns:
                continue

            if col2 not in df.columns:
                continue

            c1 = df[col1]
            c2 = df[col2]

            # remove if columns are exact duplicates
            if (c1 != c2).sum() == 0:
                df.rename(columns={col1: col_new}, inplace=True)
                del df[col2]
                continue

            # see if the differences are when one column is NA and the other is not
            # (can happen on a left-join, for example)
            c1b = np.where(c1.isna(), c2, c1)
            c2b = np.where(c2.isna(), c1, c2)

            if (c1b != c2b).sum() == 0:
                del df[col1]
                del df[col2]
                df[col_new] = c1b
                continue

            if drop_diff_cols:
                del df[col1]
                del df[col2]
                continue

    return df


def x_match(self, values, exclude=False):
    """
    Filter DataFrame by column values

    Examples:
    >> df['animal'].x_match(['cat', 'dog', '*ouse'], exclude=True)
    """

    possible_values = self.unique()

    if isinstance(values, str):
        values = [values]

    keep_exact = []
    for keep_pattern in values:
        if isinstance(keep_pattern, str):
            vals = [v for v in possible_values if fnmatch.fnmatch(v, keep_pattern)]
            keep_exact.extend(vals)

        else:
            keep_exact.append(keep_pattern)

    idxs = self.isin(keep_exact)
    if exclude:
        idxs = ~idxs

    return idxs


def x_filter_by(self, values, col_name, exclude=False, dropna=True, ignore_index=None, sanity_checks=None):
    """
    Filter DataFrame by column values

    Examples:
    >> df.x_filter_by(['cat', 'dog', '*ouse'], 'animal', exclude=True)
    """

    ignore_index = xsettings.get_default(xsettings.IGNORE_INDEX, ignore_index)
    sanity_checks = xsettings.get_default(xsettings.SANITY_CHECKS, sanity_checks)

    if dropna:
        self = self.dropna(subset=[col_name])

    idxs = x_match(self[col_name], values, exclude=exclude)

    df_filtered = self[idxs]

    if ignore_index:
        df_filtered = df_filtered.reset_index(drop=True)

    if sanity_checks:
        assert len(df_filtered) > 0, f"attr_name={col_name}, keep_values={values}, exclude={exclude}"

    return df_filtered


def x_split_on(self, values, col_name, ignore_index=None, drop_key_col=False):
    """
    Filters & splits dataframe by values

    >> df_cats, df_dogs = df.x_split_on(['cat', 'dog'], 'animal')
    """
    ignore_index = xsettings.get_default(xsettings.IGNORE_INDEX, ignore_index)

    res = []
    for value in values:
        df_value = x_filter_by(self, [value], col_name, ignore_index=ignore_index)
        if drop_key_col:
            del df_value[col_name]

        res.append(df_value)

    return tuple(res)


def x_append(self, other, ignore_index=None, verify_integrity=False, sort=False):
    ignore_index = xsettings.get_default(xsettings.IGNORE_INDEX, ignore_index)

    if self is None:
        return other

    if other is None:
        return self

    return self.append(other, ignore_index=ignore_index, verify_integrity=verify_integrity, sort=sort)


def x_replace(self, valid_vals=tuple(), replace_vals=None, default=None):
    replace_vals = replace_vals or dict()
    valid_vals = tuple(set(valid_vals) | set(replace_vals.keys()) | set(replace_vals.values()))

    def do(x):
        if x not in valid_vals:
            return default

        return replace_vals.get(x, x)

    sa_replaced = self.apply(do)
    return sa_replaced


def x_clean_text(self):
    return self.apply(lambda x: slugify(x, separator="_"))


def x_as_float(self, nan_texts=None):
    nan_texts = xsettings.get_default(xsettings.NAN_TEXTS, nan_texts)

    def conv(x):
        try:
            return np.float(x)
        except ValueError:
            if x.strip() in nan_texts:
                return np.nan
            raise ValueError(x)

    return self.apply(conv)


def x_long_to_wide(df, key_cols, widen_on, value_cols, include_other_cols=True, sep='_'):
    df_res = pd.pivot(df, index=key_cols, columns=widen_on, values=value_cols)
    new_cols = [sep.join([str(a) for a in i]) for i in df_res.columns.to_flat_index()]
    df_res.columns = new_cols
    df_res.reset_index(inplace=True)

    if include_other_cols:
        df2 = df.drop_duplicates(key_cols)
        df_res = df_res.merge(df2, on=key_cols, how='left')
        col_diff = set(df_res.columns) - set(df.columns)
        cols = df.columns.to_list() + [c for c in df_res.columns if c in col_diff]
        df_res = df_res[cols].copy()

    return df_res


def x_apply_on_group(df, by, func, include_other_cols=True):
    """
    apply a function to groups in dataframe
    :param df: input df
    :param by: columns to group on
    :param func: accepts df, can return dict (multiple columns)
    :param include_other_cols: if True, will include other cols in df, not just those specified in 'by'
           (TODO: add check (is_unique) to make sure only add columns with unique values per group)
    :return: a dataframe, where each row has unique 'by' + extra cols

    >>> x_apply_on_group(df, ['animal', 'style'], lambda dfx: return {'max_height': dfx.height.max()})
    """

    def wrapper(dfx):
        res = func(dfx)
        if isinstance(res, dict):
            res = pd.Series(res)

        return res

    g = df.groupby(by)
    df_res = g.apply(wrapper)
    df_res = df_res.reset_index()

    if include_other_cols:
        df2 = df.drop_duplicates(by)
        df_res = df_res.merge(df2, on=by, how='left')
        cols = df.columns.to_list() + sorted(set(df_res.columns) - set(df.columns))
        df_res = df_res[cols].copy()

    return df_res


def x_groupby(self, by, aggs, ignore_index=None, **kwargs):
    """
    A simpler groupby interface
    (inspired by Turicreate)

    >> df.x_groupby('a', {'min_b': xagg.Min('b'), 'max_c': xagg.Max('c'), 'n': xagg.Count()})
    """

    ignore_index = xsettings.get_default(xsettings.IGNORE_INDEX, ignore_index)

    d = dict()
    for k, v in aggs.items():
        if isinstance(v, xagg._Agg):
            tup = v.as_tuple()
            if tup[0] is None:
                tup = tuple([self.columns[0], tup[1]])
            d[k] = tup

        elif isinstance(v, tuple):
            d[k] = v

        else:
            raise ValueError(f"{k}: {v}")

    g = self.groupby(by=by, **kwargs).agg(**d)

    if ignore_index:
        g = g.reset_index()

    return g


def x_calc_rank_num(self, key_col_name, score_col_name, ascending=True):
    """
    creates a series ranking each group of key_col_name by score_col_name

    >> df.x_calc_rank_num('user_id', 'product_score', ascending=False)
    """

    return self.groupby(key_col_name)[score_col_name].rank(ascending=ascending).astype(int)


def x_add_history(self, key, value):
    if not hasattr(self, '_x_history'):
        self._metadata.append('_x_history')
        self._x_history = []

    self._x_history.append((key, value))


def x_get_history(self):
    if not hasattr(self, '_x_history'):
        return []

    return self._x_history


def x_set_data_type(self, data_type):
    """
    sets logical data type: binary, nominal, ordinal, interval, ratio, temporal, geojson
    (inspired by altair, etc.)
    what about cyclical? (can be either ordinal [day of week] or interval [hour of day]
    """

    if not hasattr(self, '_x_data_type'):
        self._metadata.append('_x_data_type')

    data_type = data_type.lower()[0]
    data_type = {
        'b': 'binary',
        'n': 'nominal',
        'o': 'ordinal',
        'i': 'interval',
        'r': 'ratio',
        'q': 'ratio',
        't': 'temporal',
        'g': 'geojson'
    }[data_type]

    self._x_data_type = data_type


def x_get_data_type(self):
    if not hasattr(self, '_x_data_type'):
        return None

    return self._x_data_type


def x_set_column_type(self, column_type):
    """
    sets logical column type: target, feature, meta, key
    - a target variable
    - a feature (can be used for training)
    - meta data (not to use as a training feature, but might be confounding)
    - a sample key (should be unique)
    """

    if not hasattr(self, '_x_column_type'):
        self._metadata.append('_x_column_type')

    column_type = column_type.lower()[0]
    column_type = {
        't': 'target',
        'f': 'feature',
        'm': 'meta',
        'k': 'key'
    }[column_type]

    self._x_column_type = column_type


def x_get_column_type(self):
    if not hasattr(self, '_x_column_type'):
        return None

    return self._x_column_type


def x_rename(self, mapper=None, index=None, columns=None, axis=None, copy=True, inplace=False, level=None, errors='raise'):
    if columns is not None:
        for name_from, name_to in columns.items():
            xsettings.x_add_desc(name_to, name_from)

    self_orig = self
    self = self.rename(mapper=mapper, index=index, columns=columns, axis=axis, copy=copy, inplace=inplace, level=level, errors=errors)

    self = self if self is not None else self_orig
    return self


def x_sort_on_lookup(self: pd.DataFrame, on, sa_lookup: pd.Series, ascending=True):
    """
    Given lookup Series, merge it based 'on', and then sort, and remove
    df.x_sort_on_lookup('height', df.groupby('height')['width'].mean())
    """
    df = self.copy()
    df['tmp_x_sort_on_lookup'] = df.apply(lambda r: sa_lookup[r[on]], axis=1)
    df = df.sort_values('tmp_x_sort_on_lookup', ascending=ascending)
    del df['tmp_x_sort_on_lookup']
    return df


def x_clean_column_names(self, inplace=False, remove_parens="auto"):
    """
    renames columns to pythonable names, " Hi there" --> "hi_there"
    """

    orig_columns = self.columns
    new_columns = orig_columns[:]
    if remove_parens:
        new_columns = [c.split('(')[0] for c in new_columns]
        new_columns = [c.split('[')[0] for c in new_columns]

    new_columns = np.array([snakecase(c) for c in new_columns])

    a = pd.Series(new_columns)
    counts = a.value_counts()
    counts = counts[counts > 1]
    bad_keys = counts.index.values

    new_columns = np.where(~np.isin(new_columns, bad_keys), new_columns, np.array([slugify(c, separator='_') for c in orig_columns]))

    a = pd.Series(new_columns)
    counts = a.value_counts()

    used_counts = Counter()

    def get_new_name(col):
        if counts[col] == 1:
            return col

        used_counts[col] += 1
        c = used_counts[col]
        return f"{col}_{c}"

    new2 = [get_new_name(c) for c in new_columns]
    assert len(set(new2)) == len(orig_columns), "not all column names become unique"

    rename_dict = dict(zip(orig_columns, new2))
    return x_rename(self, columns=rename_dict, inplace=inplace)


def x_apply_multi(self):
    # TODO: do
    return self


def x_drop_by_counts(self, count_on, keep_count=None, min_count=None, max_count=None):
    """
    count number of unique values for `count_on`, and filter df by `keep_count`
    """

    sa_counts = self[count_on].value_counts()
    if keep_count:
        sa_counts = sa_counts[sa_counts == keep_count]
    if min_count:
        sa_counts = sa_counts[sa_counts >= min_count]
    if max_count:
        sa_counts = sa_counts[sa_counts <= min_count]

    self = x_filter_by(self, sa_counts.index.values, count_on)
    return self


def x_timedelta_as_hours(self):
    """
    Given a Series containing Timedelta, converts it to Series containing hours (float)
    """
    return self.apply(lambda d: d.days * 24 + d.seconds / (60 * 60))


def x_add_uid(self, on, new_col, start_with=1, sort=True):
    """
    Given a df, converts to unique IDs
    """
    if isinstance(on, str):
        on = [on]

    df = self.copy()
    new_col_temp = None
    if new_col in df.columns:
        new_col_temp = f"{new_col}__tmp"
        df.rename(columns={new_col: new_col_temp}, inplace=True)

        if new_col in on:
            on.remove(new_col)
            on.append(new_col_temp)

    df2 = df[on]
    df2 = df2.drop_duplicates()
    if sort:
        df2 = df2.sort_values(on)

    df2 = df2.reset_index(drop=True)
    df2 = df2.reset_index()
    df2.rename(columns={'index': new_col}, inplace=True)
    df2[new_col] = df2[new_col] + start_with
    df_new = df.merge(df2, on=on)

    if new_col_temp:
        del df_new[new_col_temp]

    return df_new


def x_undo_dummies(df, prefix_sep="_"):
    """
    Given df with dummy columns (created using pd.get_dummies()), undo the dummies
    Credits: https://newbedev.com/reverse-a-get-dummies-encoding-in-pandas
    """
    cols2collapse = {
        item.split(prefix_sep)[0]: (prefix_sep in item) for item in df.columns
    }
    series_list = []
    for col, needs_to_collapse in cols2collapse.items():
        if needs_to_collapse:
            undummified = (
                df.filter(like=col)
                .idxmax(axis=1)
                .apply(lambda x: x.split(prefix_sep, maxsplit=1)[1])
                .rename(col)
            )
            series_list.append(undummified)
        else:
            series_list.append(df[col])
    undummified_df = pd.concat(series_list, axis=1)
    return undummified_df


def monkey_patch():
    pandas.Series.x_match = x_match
    pandas.Series.x_replace = x_replace
    pandas.Series.x_clean_text = x_clean_text
    pandas.Series.x_add_history = x_add_history
    pandas.Series.x_get_history = x_get_history
    pandas.Series.x_set_data_type = x_set_data_type
    pandas.Series.x_get_data_type = x_get_data_type
    pandas.Series.x_set_column_type = x_set_column_type
    pandas.Series.x_get_column_type = x_get_column_type
    pandas.Series.x_as_hours = x_timedelta_as_hours

    pandas.DataFrame.x_filter_by = x_filter_by
    pandas.DataFrame.x_split_on = x_split_on
    pandas.DataFrame.x_append = x_append
    pandas.DataFrame.x_groupby = x_groupby
    pandas.DataFrame.x_calc_rank_num = x_calc_rank_num
    pandas.DataFrame.x_add_history = x_add_history
    pandas.DataFrame.x_get_history = x_get_history
    pandas.DataFrame.x_clean_column_names = x_clean_column_names
    pandas.DataFrame.x_rename = x_rename
    pandas.DataFrame.x_sort_on_lookup = x_sort_on_lookup
    pandas.DataFrame.x_undo_dummies = x_undo_dummies







import pandas as pd
import numpy as np
from slugify import slugify
from munch import Munch as Dict
from . import xsettings, xpd


def drop_na_cols_df(df):
    if len(df) == 0:
        return df

    df = df.copy()
    for col_name in df.columns:
        sa = df[col_name]

        if not pd.api.types.is_numeric_dtype(sa):
            continue

        if np.isnan(sa).sum() == len(sa):
            del df[col_name]

    return df


def read_csv(csv_path, na_values=None, low_memory=False, drop_na_cols=True, skiprows=None, skipfooter=0, dtype=None, rtl_cols=tuple()):
    """
    A simple wrapper for read_csv that cleans it up a bit in the process..
    """
    na_values = xsettings.get_default(xsettings.NAN_TEXTS, na_values)

    engine = 'python' if skipfooter else 'c'

    df = pd.read_csv(csv_path, keep_default_na=False, na_values=na_values, low_memory=low_memory, skiprows=skiprows, skipfooter=skipfooter, engine=engine, dtype=dtype)
    xpd.x_clean_column_names(df, inplace=True)

    if drop_na_cols:
        df = drop_na_cols_df(df)

    for col_name in rtl_cols:
        if col_name in df.columns:
            df[f"{col_name}_rtl"] = df[col_name].apply(lambda s: s[::-1])

    return df


def read_excel(excel_path, na_values=None, drop_na_cols=True):
    """
    A simple wrapper for read_excel that cleans it up a bit in the process,
      as well as returns a different dataframe for each sheet
    """
    na_values = xsettings.get_default(xsettings.NAN_TEXTS, na_values)

    res = pd.read_excel(excel_path, sheet_name=None, na_values=na_values)

    res_new = Dict()
    for k in list(res.keys()):
        k_new = slugify(k, separator='_')
        df = res[k]
        xpd.x_clean_column_names(df, inplace=True)
        if drop_na_cols:
            df = drop_na_cols_df(df)

        assert k_new not in res_new, k_new
        res_new[k_new] = df

    return res_new


def filter_all_on(*args, on=None):
    """
    Given two or more dataframes, returns them filtered as if they went through an inner-join
    """

    assert on, "must provide 'on' parameter"
    if len(args) < 2:
        return args

    set_all = args[0][on].unique()
    for df in args[1:]:
        set_curr = df[on].unique()
        set_all = np.intersect1d(set_all, set_curr)

    filtered = []
    for df in args:
        df = df[df[on].isin(set_all)]
        filtered.append(df)

    return tuple(filtered)




import networkx as nx
import pandas as pd
import numpy as np
from sklearn import decomposition, manifold
umap = None


def x_reduce_dim_2d(df, method='umap', **kwargs):
    """
    Various ways of reducing to 2D
    Note: can pass df.T to get feature similarity
    """

    if not isinstance(df, pd.DataFrame):
        df = pd.DataFrame(df)

    if method == 'spring':
        G = nx.Graph(df.T.corr())
        pos = nx.spring_layout(G)
        df_2d = pd.DataFrame(pos).T

    elif method == 'umap':
        global umap
        if umap is None:
            import umap

        pos = umap.UMAP(**kwargs).fit_transform(df)
        df_2d = pd.DataFrame(pos)

    elif method == 'tsne':
        pos = manifold.TSNE(**kwargs).fit_transform(df)
        df_2d = pd.DataFrame(pos)

    elif method == 'pca':
        pos = decomposition.PCA(n_components=2, **kwargs).fit_transform(df)
        df_2d = pd.DataFrame(pos)

    else:
        raise ValueError(method)

    df_2d.columns = ['x', 'y']
    return df_2d


def x_inspect_cols(df):
    """
    Go over columns, get general idea of what they contain
    """

    for col_name in df.columns:
        sa_orig = df[col_name]
        sa = sa_orig.dropna()
        num_unique = len(sa.unique())
        topk = (sa.value_counts() / len(sa)).iloc[:5]
        topk_text = ", ".join([ f"{k}={100*v:.2f}%" for k,v in topk.items()])
        prec_na = 1 - len(sa)/len(sa_orig)
        print(f"{col_name}: dtype={sa.dtype} na={100*prec_na:.1f}% #unique={num_unique} top5={topk_text}")


def x_corr_with_target(df, target):
    """
    Go over columns, see how well they correlate with target
    """

    df, new_target = x_prep_col(df, target, on_na='drop')
    assert len(new_target) == 1, new_target
    if target != new_target:
        print(f'New Target: {target} ==> {new_target}')

    target = new_target[0]

    rows = []
    for col_name in df.columns:
        if col_name == target:
            continue

        df_sub = df[[col_name, target]].dropna()
        df_sub, new_cols = x_prep_col(df_sub, col_name)
        if len(new_cols) == 0:
            print(f"{col_name}: SKIPPED")
            continue

        elif len(new_cols) == 1:
            new_col = new_cols[0]
            corr = np.corrcoef(df_sub[new_col], df_sub[target])[0, 1]
            print(f"{new_col}: {corr}")
            rows.append({'orig_col': col_name, 'new_col': new_col, 'corr': corr, 'abs_corr': abs(corr)})

        else:
            print(f"{col_name}:")
            for new_col in new_cols:
                corr = np.corrcoef(df_sub[new_col], df_sub[target])[0, 1]
                print(f"   {new_col}: {corr}")
                rows.append({'orig_col': col_name, 'new_col': new_col, 'corr': corr, 'abs_corr': abs(corr)})

    df_corr = pd.DataFrame(rows)
    df_corr = df_corr.sort_values('abs_corr', ascending=False).reset_index(drop=True)
    return df_corr


def x_prep_df(df, target=None, on_na='auto'):
    """
    Go over columns, prepare for modeling (fill NA, dummy vars, etc.)
    NOTE: drops NA for target col, does "auto" for others
    """
    df = df.copy()

    new_target = None
    if target:
        df, new_target = x_prep_col(df, target, on_na='drop', inplace=True)
        assert len(new_target) == 1
        new_target = new_target[0]

    for col_name in df.columns:
        if col_name == new_target:
            continue

        df, _ = x_prep_col(df, col_name, on_na=on_na)

    if target is None:
        return df

    return df, new_target


def x_prep_col(df, col_name, on_na='auto', inplace=False):
    """
    prepare a column for modeling (fill NA, dummy vars, etc)
    """

    df = df.copy()
    new_cols = [col_name]

    if on_na == 'auto':
        if pd.api.types.is_float_dtype(df[col_name]):
            on_na = 'mean'
        elif pd.api.types.is_numeric_dtype(df[col_name]):
            on_na = 'median'
        else:
            on_na = 'na_value'

    if on_na == 'drop':
        df = df.dropna(subset=[col_name])

    elif on_na == 'mode':
        mode = df[col_name].value_counts().index.values[0]
        df[col_name] = np.where(df[col_name].isna(), mode, df[col_name])

    elif on_na == 'mean':
        mean = df[col_name].mean()
        df[col_name] = np.where(df[col_name].isna(), mean, df[col_name])

    elif on_na == 'median':
        median = df[col_name].median()
        df[col_name] = np.where(df[col_name].isna(), median, df[col_name])

    elif on_na == 'na_value':
        df[col_name] = np.where(df[col_name].isna(), 'NA', df[col_name])

    else:
        raise ValueError(on_na)

    if not pd.api.types.is_string_dtype(df[col_name]):
        return df, new_cols

    sa_col = df[col_name]
    unique_vals = np.unique(sa_col)
    num_unique = len(unique_vals)
    if num_unique == 1:
        del df[col_name]
        new_cols = []

    elif num_unique == 2:
        val_sel = sa_col.value_counts().index.values[0]
        new_val = (sa_col == val_sel).astype(int)
        if inplace:
            df[col_name] = new_val

        else:
            new_col_name = f"{col_name}_{val_sel}"
            new_cols = [new_col_name]
            df[new_col_name] = new_val
            del df[col_name]

    else:
        assert inplace is False, f"{col_name}: can't do inplace with multiple categorical values ({num_unique} unique)"

        new_cols = []
        for val_sel in unique_vals:
            new_val = (sa_col == val_sel).astype(int)
            new_col_name = f"{col_name}_{val_sel}"
            new_cols.append(new_col_name)
            df[new_col_name] = new_val

        del df[col_name]

    return df, new_cols



import numpy


class _Agg:
    def __init__(self, col_name):
        self.col_name = col_name

    def _get_func(self):
        raise NotImplemented

    def as_tuple(self):
        func = self._get_func()
        return self.col_name, func


class Count(_Agg):
    def __init__(self):
        super().__init__(None)

    def _get_func(self):
        return numpy.size


class Min(_Agg):
    def _get_func(self):
        return numpy.min


class Max(_Agg):
    def _get_func(self):
        return numpy.max


class Mean(_Agg):
    def _get_func(self):
        return numpy.mean


class Std(_Agg):
    def _get_func(self):
        return numpy.std


class Any(_Agg):
    def _get_func(self):
        return lambda x: x.values[0] if len(x) else numpy.NaN


class Lambda(_Agg):
    def __init__(self, col_name, func):
        super().__init__(col_name)
        self.func = func

    def _get_func(self):
        return self.func





import unittest
from ..xnp import *
from ..xchecks import *

monkey_patch()


class TestXNP(unittest.TestCase):
    def test_x_sum(self):
        a = numpy.array([2, 4, 6])
        self.assertEqual(a.sum(), numpy.x_sum(a))

    def test_x_dot(self):
        a = numpy.array([2, 4, 6]).reshape((3, 1))
        a2 = a.reshape((1, 3))
        check_equals(numpy.dot(a, a2), numpy.x_dot(a, a2))





import numpy as np


def x_calc_gini(a, presorted=False):
    assert sum(a<0) == 0, "can't have neg values"
    if not presorted:
        a = np.dropna(a)
        np.sort(a)

    total = 0
    for i, xi in enumerate(a[:-1], 1):
        total += np.sum(np.abs(xi - a[i:]))
    return total / (len(a)**2 * np.mean(a))



"""

"""

__version__ = "0.1.0"
__author__ = 'Ido Carmi'
__credits__ = 'Hermetric Software Services Ltd.'


# import pandas_sets
from . import xagg, xcache, xchecks, xmunge, xnp, xpd, xplt


def monkey_patch():
    xpd.monkey_patch()
    xnp.monkey_patch()




"""
Factors Analysis
"""
import numpy as np
from sklearn import model_selection, ensemble
from tqdm import tqdm


class BinaryFactorsAnalysis:
    def __init__(self, clf):
        self.clf = clf

    def fit(self, X):
        self.clf.fit(X)

    def get_W_H(self, X):
        W = self.clf.transform(X)  # rows
        H = self.clf.components_  # cols
        return W, H

    def predict_proba(self, X):
        W, H = self.get_W_H(X)
        X_hat = np.dot(W, H)
        return X_hat

    def predict(self, X, idx_groups=None):
        X_hat = self.predict_proba(X)
        X_pred = np.zeros(shape=X_hat.shape)
        for idx_start, idx_end in idx_groups:
            idx_end += 1

            argmax = X_hat[:, idx_start:idx_end].argmax(axis=1)
            X_pred[:, idx_start:idx_end][np.arange(X_pred.shape[0]), argmax] = 1

        return X_pred


def factors_cv(X, create_clf, n_splits=12, how='predict', idx_groups=None):
    kfold = model_selection.KFold(n_splits=n_splits)
    X_test_hat = np.zeros(shape=X.shape)

    for train_index, test_index in tqdm(kfold.split(X), total=n_splits):
        X_train = X[train_index]
        X_test = X[test_index]

        clf2 = create_clf()
        clf2.fit(X_train)

        f = getattr(clf2, how)
        if idx_groups:
            X_hat = f(X_test, idx_groups=idx_groups)
        else:
            X_hat = f(X_test)

        X_test_hat[test_index] = X_hat

    return X_test_hat



from collections import namedtuple
import numpy as np
from scipy import stats
from sklearn import metrics

ModelPredStats = namedtuple("ModelPredStats", ["r2", "r2_adj", "p_value", "mae", "rmse", "p_value_err_normal"])


def x_model_pred_stats(y_true, y_pred, y_alt='mean', k=None):
    """
    Calculates various statistics on model predictions, including:
        p_value: the p_value of the test that the absolute errors of the pred are less than the absolute errors of the alt
        p_value_err_normal: the p_value of the test that the errors are normally distributed
        r2_adj: adjusted r-squared (requires the k parameter to be set)

    Args:
        y_true: true values
        y_pred: predicted values
        y_alt: alternative (null) model, typically mean of train data
        k: number of variables in the model (for calculating r2_adj)

    Returns: ModelPredStats
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)

    if y_alt == 'mean':
        y_alt = y_true.mean()

    err_pred = np.abs(y_true - y_pred)
    err_alt = np.abs(y_true - y_alt)

    res = stats.ttest_rel(err_pred, err_alt, alternative='less')
    p_value = res.pvalue
    if p_value is np.nan:
        p_value = None

    errors = y_true - y_pred
    p_value_err_normal = None
    try:
        p_value_err_normal = stats.normaltest(errors).pvalue
    except ValueError:
        pass

    if p_value_err_normal is np.nan:
        p_value_err_normal = None

    r2 = metrics.r2_score(y_true, y_pred)
    n = len(y_true)
    r2_adj = None
    if k is not None and n - k - 1 > 0:
        r2_adj = 1 - ((1-r2)*(n-1))/(n-k-1)

    mae = np.mean(err_pred)
    rmse = metrics.mean_squared_error(y_true, y_pred, squared=False)
    res = ModelPredStats(r2=r2, r2_adj=r2_adj, p_value=p_value, mae=mae, rmse=rmse, p_value_err_normal=p_value_err_normal)
    return res



from . import xpd, xnp, xplt


def x_monkey_patch():
    xpd.monkey_patch()
    xnp.monkey_patch()
    xplt.monkey_patch()


def split_X_y(df, target):
    df = df.copy()
    y = df[target]
    del df[target]
    return df, y



import numpy as np
from sklearn.base import clone, BaseEstimator
from sklearn import metrics


class OrdinalClassifier(BaseEstimator):
    """
    Based on classifier that returns binary probability, builds a bunch of classifiers that together solve the ordinal classifier problem.

    Credits:
      https://www.cs.waikato.ac.nz/~eibe/pubs/ordinal_tech_report.pdf
      https://stackoverflow.com/questions/57561189/multi-class-multi-label-ordinal-classification-with-sklearn
    """

    def __init__(self, clf):
        """
        :param clf: binary classifier with 'predict_proba()'
        """

        self.clf = clf
        self.clfs = {}
        self.unique_class = None

    def fit(self, X, y):
        self.unique_class = np.sort(np.unique(y))
        assert self.unique_class.shape[0] > 2, 'looks like a binary problem'

        for i in range(self.unique_class.shape[0]-1):
            # for each k - 1 ordinal value we fit a binary classification problem
            binary_y = (y > self.unique_class[i]).astype(np.uint8)
            clf = clone(self.clf)
            clf.fit(X, binary_y)
            self.clfs[i] = clf

    def predict_proba(self, X):
        clfs_predict = {k: self.clfs[k].predict_proba(X) for k in self.clfs}
        predicted = []
        for i, y in enumerate(self.unique_class):
            if i == 0:
                # V1 = 1 - Pr(y > V1)
                predicted.append(1 - clfs_predict[i][:,1])
            elif i in clfs_predict:
                # Vi = Pr(y > Vi-1) - Pr(y > Vi)
                 predicted.append(clfs_predict[i-1][:,1] - clfs_predict[i][:,1])
            else:
                # Vk = Pr(y > Vk-1)
                predicted.append(clfs_predict[i-1][:,1])
        return np.vstack(predicted).T

    def predict(self, X):
        return np.argmax(self.predict_proba(X), axis=1)

    def score(self, X, y, sample_weight=None):
        _, indexed_y = np.unique(y, return_inverse=True)
        return metrics.accuracy_score(indexed_y, self.predict(X), sample_weight=sample_weight)



import pickle
from hashlib import md5
from scriptine import path

from .xcache import x_cached_call

"""
https://github.com/stan-dev/pystan2
https://pystan2.readthedocs.io/en/latest/
"""


def cached(model_code=None, model_path=None, model_name=None, cache_folder='/tmp', **kwargs):
    """Use just as you would `stan`"""
    import pystan

    assert model_code or model_path
    cache_folder = path(cache_folder)
    if model_path:
        model_path = path(model_path).abspath()
        print(f'Stan model: {model_path}')
        with open(model_path, "r") as f:
            model_code = f.read()

    code_hash = md5(model_code.encode('ascii')).hexdigest()
    if model_name is None:
        cache_fn = 'cached-model-{}.pkl'.format(code_hash)
    else:
        cache_fn = 'cached-{}-{}.pkl'.format(model_name, code_hash)

    cache_fn = cache_folder.joinpath(cache_fn)
    try:
        sm = pickle.load(open(cache_fn, 'rb'))
    except:
        sm = pystan.StanModel(model_code=model_code, **kwargs)
        with open(cache_fn, 'wb') as f:
            pickle.dump(sm, f)
    else:
        print("Using cached StanModel")
    return sm



import datetime as dt
import pandas as pd
import seaborn as sns
import numpy as np
import arviz as az

from sklearn import linear_model
from sklearn import metrics
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import matplotlib.ticker as mticker
from ds_utils.metrics import plot_confusion_matrix as _plot_confusion_matrix, visualize_accuracy_grouped_by_probability
from scriptine import path

from . import xproblem, xpd, xsettings, xcalc, xstats


def plot_gini(a, num_bins=101, xlabel='samples', ylabel='values', title='', figsize=(8,8)):
    """
    Credit: https://stackoverflow.com/questions/39512260/calculating-gini-coefficient-in-python-numpy
    """
    a = pd.Series(a)
    a = a.dropna()
    np.sort(a)
    gini_val = xcalc.x_calc_gini(a, presorted=True)

    def G(v):
        bins = np.linspace(0., 100., num_bins)
        total = float(np.sum(v))
        yvals = []
        for b in bins:
            bin_vals = v[v <= np.percentile(v, b)]
            bin_fraction = (np.sum(bin_vals) / total) * 100.0
            yvals.append(bin_fraction)

        return bins, yvals

    bins, result = G(a)
    plt.figure(figsize=figsize)
    # plt.subplot(2, 1, 1)
    plt.plot(bins, result, label="observed")
    plt.plot(bins, bins, '--', label="perfect eq.")
    plt.xlabel(f"fraction of {xlabel}")
    plt.ylabel(f"fraction of {ylabel}")
    title2 = f"GINI={gini_val:.4f}"
    if title:
        title2 = f"{title}: {title2}"
    plt.title(title2)
    plt.legend()
    # plt.subplot(2, 1, 2)
    # plt.hist(a, bins=20)
    plt.tight_layout()
    plt.show()


def plot_feature_importances(folds, title='', top_k=None):
    df = xproblem.calc_feature_importances(folds, flat=True)
    if df is None:
        return

    fis = df.groupby('feature_name')['feature_importance'].mean()
    if top_k:
        fis = fis.sort_values(ascending=False)[:top_k]
        df = df[df.feature_name.isin(fis.index.values)]

    df = xpd.x_sort_on_lookup(df, 'feature_name', fis, ascending=True)
    sns.catplot(data=df, y='feature_name', x='feature_importance')
    plt.xlim([0, None])
    if title:
        plt.title(title)

    plt.tight_layout()
    plt.show()

    return


def plot_roc_curve(y_true, y_score, title=''):
    auc = metrics.roc_auc_score(y_true, y_score)
    fper, tper, thresholds = metrics.roc_curve(y_true, y_score)
    plt.plot(fper, tper, color='orange', label='ROC')
    plt.plot([0, 1], [0, 1], color='darkblue', linestyle='--')
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')

    title2 = f'ROC Curve (AUC={auc:.3f})'
    if title:
        title2 = f"{title}: {title2}"

    plt.title(title2)
    plt.legend()
    plt.tight_layout()
    plt.show()


def plot_confusion_matrix(y_true, y_pred, y_score=None, labels=None, title=''):
    auc = metrics.roc_auc_score(y_true, y_score) if y_score is not None else None

    y_true = pd.Series(y_true)
    counts = sorted([(l, c) for l, c in y_true.value_counts().items()])
    counts_str = ", ".join([f"{l}={c}" for l, c in counts])

    labels = labels or sorted(y_true.unique())

    _plot_confusion_matrix(y_true, y_pred, labels=labels, cbar=False)

    title2 = f"Counts: {counts_str} (total={len(y_true)})"
    if auc is not None:
        title2 = f"{title2} AUC={auc:.3f}"
    if title:
        title2 = f"{title}: {title2}"

    plt.title(title2)

    plt.tight_layout()
    plt.show()


def plot_model_scores(y_true, y_score, bins=25, title='', show=True):
    """
    Useful of comparing model scores for the different targets
    """

    df = pd.DataFrame({'Target': y_true, 'Model Score': y_score})
    sns.histplot(data=df, x='Model Score', hue='Target', element="step", common_norm=False, stat='percent', bins=bins)

    title2 = 'Histogram of model scores'
    if title:
        title2 = f"{title}: {title2}"

    plt.title(title2)
    plt.tight_layout()

    if show:
        plt.show()


def plot_model_accuracy_binned(df, title=''):
    df = df.copy()
    df['correct'] = df.pred == df.target
    df['bin'] = df.prob_1.apply(lambda p: float(str(p)[:3]))
    g = df.groupby('bin')
    df_bins = g.mean()['correct'].reset_index()
    plt.bar(df_bins['bin'], df_bins['correct'], align='edge', width=0.08)

    title2 = 'Model accuracy by score (binned)'
    if title:
        title2 = f"{title}: {title2}"

    plt.title(title2)

    plt.tight_layout()
    plt.show()

def plot_score_comparison(scores: dict, key_label='Dataset', score_label='Model Score', bins=25, title=''):
    """
    Useful for comparing the scores of various datasets:
    > plot_score_comparison({'train': train_scores, 'test': test_scores, 'blind': blind_scores)
    """

    rows = []
    for k, scores in scores.items():
        for score in scores:
            rows.append({key_label: k, score_label: score})

    df = pd.DataFrame(rows)
    sns.histplot(data=df, x=score_label, hue=key_label, element="step", common_norm=False, stat='percent', bins=bins)

    title2 = f'Histogram of model scores per {key_label}'
    if title:
        title2 = f"{title}: {title2}"

    plt.title(title2)
    plt.tight_layout()
    plt.show()


def plot_model_scores_ratios(y_true, y_score, bins=25, ratio_of=1):
    df = pd.DataFrame({'target': y_true, 'score': y_score})
    s_min = y_score.min()
    s_max = y_score.max()
    s_range = s_max - s_min
    borders = np.linspace(s_min, s_max+s_range*.0001, bins+1)

    rows = []
    for s_start, s_end in zip(borders[:-1], borders[1:]):
        s_mid = (s_start + s_end) / 2
        df_g = df[(df.score >= s_start) & (df.score < s_end)]
        if len(df_g) == 0:
            continue

        r = (df_g.target == ratio_of).sum() / len(df_g)
        rows.append({'s_start': df_g.score.min(), 's_end': df_g.score.max(), 'ratio': r})

    df_rows = pd.DataFrame(rows)

    for row in df_rows.itertuples():
        plt.plot([row.s_start, row.s_end], [row.ratio, row.ratio], color='black')

    plt.title('Histogram of model scores')
    plt.tight_layout()
    plt.show()


def plot_corr_heatmap(df, title='Correlation Heatmap', fontsize=12, pad=12, cmap='BrBG', figsize=(15,15)):
    """
    Credits: https://medium.com/@szabo.bibor/how-to-create-a-seaborn-correlation-heatmap-in-python-834c0686b88e
    """

    plt.subplots(figsize=figsize)
    df_corr = df.corr()
    mask = np.triu(np.ones_like(df_corr, dtype=np.bool))

    hm = sns.heatmap(df_corr, vmin=-1, vmax=1, annot=True, cmap=cmap, mask=mask)
    hm.set_title(title, fontdict={'fontsize': fontsize}, pad=pad)
    plt.tight_layout()
    plt.show()


def plot_pie(vals=None, counts=None, title=''):
    if vals is not None:
        vals = pd.Series(vals)
        counts = vals.value_counts()

    assert counts is not None, "either vals or counts must be provided"

    plt.figure(figsize=(6, 6))
    plt.pie(counts.values, labels=counts.index.values, autopct='%1.1f%%', shadow=False, startangle=0)

    title2 = f"total count={counts.sum()}"
    if title:
        title2 = f"{title}: {title2}"
    plt.title(title2)

    plt.tight_layout()
    plt.show()


def plot_counts(df, on, sort_by_counts=True, title=''):
    counts = df[on].value_counts()
    counts = counts.sort_values(ascending=False)
    plt.clf()
    title2 = f"Counts on {on}"
    if title:
        title2 = f"{title}: {title2}"
    plt.title(title2)
    x = np.arange(len(counts)) if sort_by_counts else counts.index.values
    plt.scatter(x, counts.values, s=2)
    plt.ylabel('counts')
    if sort_by_counts:
        plt.xlabel(f"{on} sorted by counts")
        plt.tick_params(
            axis='x',  # changes apply to the x-axis
            which='both',  # both major and minor ticks are affected
            bottom=False,  # ticks along the bottom edge are off
            top=False,  # ticks along the top edge are off
            labelbottom=False)  # labels along the bottom edge are off

    else:
        plt.xlabel(f"{on}")

    plt.tight_layout()
    plt.show()
    return


def big_scatter(df, x, y, colors='fire', title='', show=True, show_axis=True, figsize=(12, 12), plot_dim=(1000, 1000)):
    import datashader as ds
    import colorcet as cc

    fig, axes = plt.subplots(figsize=figsize)
    # a = ds.Canvas().points(df, 'x', 'y')
    # tf.shade(a)
    # plt.plot([-15, 15], [-15, 15], color='blue')
    # plt.show()

    cvs = ds.Canvas(plot_width=plot_dim[0], plot_height=plot_dim[1])
    agg = cvs.points(df, x, y)

    if colors == 'fire':
        img = ds.tf.set_background(ds.tf.shade(agg, cmap=cc.fire), "black")
    elif colors == 'bw':
        img = ds.tf.set_background(ds.tf.shade(agg, cmap=cc.b_cyclic_grey_15_85_c0_s25), "black")
    else:
        raise ValueError(colors)

    img2 = img.to_pil()
    plt.imshow(img2, extent=(img[x].min(), img[x].max(), img[y].min(), img[y].max()))

    if not show_axis:
        plt.axis('off')

    plt.tight_layout()

    if title:
        plt.title(title)

    if show:
        plt.show()


def post_plot(title='', legend=True, xlim=None, ylim=None, tight_layout=True, show=True, y_auc=False):
    """
    Helper function to set various plot attributes...
    """

    if y_auc:
        ylim = [0, 1]

    if title:
        plt.title(title)

    if xlim:
        plt.xlim(xlim)

    if ylim:
        plt.ylim(ylim)

    if y_auc:
        plt.axhline(y=0.5, ls=':', color='red', alpha=0.75)
        plt.axhline(y=0.65, ls=':', color='xkcd:dark yellow', alpha=0.75)
        plt.axhline(y=0.35, ls=':', color='xkcd:dark yellow', alpha=0.75)
        plt.axhline(y=0.8, ls=':', color='xkcd:green', alpha=0.75)
        plt.axhline(y=0.2, ls=':', color='xkcd:green', alpha=0.75)

    if tight_layout:
        plt.tight_layout()

    if legend:
        plt.legend()

    if show:
        plt.show()


def plot_multi(df, kind='scatter', plot_func=None, x=None, y=None, plot_on=None, group_on=None, color_on=None, label_on=None, style_on=None, size_on=None, figsize=(10,6), alpha=1.0, hdi_probs=(0.1, 0.25, 0.5, 0.8, 0.999), color_dict=None, title='', x_axis_type=None, y_axis_type=None, legend_loc='best', xlim=None, ylim=None, save_to=None, clear_folder=False, add_date=True, show=True, **kwargs):
    """

    :param df: input dataframe
    :param kind: type of plot ('scatter', 'line', 'hdi')
    :param plot_func: alternative to *kind*, can provide a custom function that takes a subset of data & plots
    :param x: name of column for x-axis
    :param y: name of column for y-axis
    :param plot_on: column name that generates different plot for each unique value
    :param group_on: column name for different subgroup of data (usually not neaded if provide color_on, etc)
    :param color_on: column name for different colors
    :param label_on: column name for different labels
    :param style_on: column name for different styles
    :param figsize:
    :param alpha: transparency
    :param hdi_probs: used for kind == 'hdi'
    :param color_dict:
    :param title:
    :param x_axis_type: can set to int
    :param y_axis_type: can be set to int
    :param legend_loc:
    :param xlim:
    :param ylim:
    :param save_to: instead of displaying, can save fig to file
    :param clear_folder:
    :param add_date: adds a date to fig
    :param kwargs: additional params that get passed to plot_func
    :return:
    """
    color_dict = color_dict or dict()

    if save_to:
        save_to = path(save_to)
        save_to.ensure_dir()

        if clear_folder:
            save_to.rmtree()
            save_to.ensure_dir()

    def to_list(i):
        if not i:
            i = []

        if isinstance(i, str):
            i = [i]

        return i

    color_on = to_list(color_on)
    style_on = to_list(style_on)
    label_on = to_list(label_on)
    group_on = to_list(group_on)

    if not label_on:
        label_on = color_on + style_on

    if not color_on:
        color_on = label_on

    if label_on:
        df = df.sort_values(label_on)

    if not group_on:
        group_on = color_on + style_on + label_on

    color_list = list(mcolors.TABLEAU_COLORS.keys()) + list(mcolors.XKCD_COLORS)
    line_styles = [
     ('solid', 'solid'),      # Same as (0, ()) or '-'
     ('dotted', 'dotted'),    # Same as (0, (1, 1)) or ':'
     ('dashed', 'dashed'),    # Same as '--'
     ('dashdot', 'dashdot'),  # Same as '-.'
     ('loosely dashed',        (0, (5, 10))),
     ('densely dashed',        (0, (5, 1))),
     ('loosely dashdotted',    (0, (3, 10, 1, 10))),
     ('dashdotted',            (0, (3, 5, 1, 5))),
     ('densely dashdotted',    (0, (3, 1, 1, 1))),
     ('dashdotdotted',         (0, (3, 5, 1, 5, 1, 5))),
     ('loosely dashdotdotted', (0, (3, 10, 1, 10, 1, 10)))]
    line_styles = [ls[1] for ls in line_styles]
    scatter_markers = list(".v1sPp*+xd|_")
    style_list = line_styles if kind == 'line' else scatter_markers
    style_param = "ls" if kind == 'line' else 'marker'

    color_id = 0
    style_id = 0
    color_lookup = dict()
    style_lookup = dict()
    for df_p, _, plot_title in xpd.x_iter_groups(df, plot_on):
        labels = set()
        plt.clf()
        fig, ax = plt.subplots(figsize=figsize)
        for df_g, _, group_title in xpd.x_iter_groups(df_p, group_on):
            color = None
            label = None
            style = None

            if label_on:
                sa_label = df_g.iloc[0][label_on]
                label = ", ".join([f"{k}={v}" for k,v in sa_label.items()])

            if label in labels:
                label = None

            labels.add(label)

            if color_on:
                sa_color_val = df_g.iloc[0][color_on]
                color_val = ", ".join([f"{k}={v}" for k, v in sa_color_val.items()])
                if color_val not in color_lookup:
                    color_lookup[color_val] = color_list[color_id]
                    color_id += 1

                color = color_lookup[color_val]

            if style_on:
                sa_style_val = df_g.iloc[0][style_on]
                style_val = ", ".join([f"{k}={v}" for k, v in sa_style_val.items()])
                if style_val not in style_lookup:
                    style_lookup[style_val] = style_list[style_id]
                    style_id += 1

                style = style_lookup[style_val]

            params = dict()
            if color is not None:
                params['color'] = color

            if style is not None:
                params[style_param] = style

            if label is not None:
                params['label'] = label

            if alpha is not None:
                params['alpha'] = alpha

            params.update(kwargs)

            if plot_func:
                plot_func(df_g, **params)

            elif kind in ['scatter', 'reg']:
                _x = df_g[x]
                _y = df_g[y]
                params['s'] = params.get('s', 0.2)
                if size_on:
                    min_s = params.get('s', 0.2)
                    sizes = df_g[size_on]
                    min_size = sizes.min()
                    max_size = sizes.max()
                    size_diff = max_size - min_size
                    size_ratio = 1
                    if size_diff > 0:
                        size_ratio = 10 / size_diff

                    sizes = (sizes - min_size) * size_ratio + min_s
                    params['s'] = sizes

                plt.scatter(x=_x, y=_y, **params)

                if kind == 'reg':
                    X = _x.to_numpy().reshape(-1, 1)
                    reg = linear_model.LinearRegression()
                    reg.fit(X, _y)
                    _y_pred = reg.predict(X)
                    st = xstats.x_model_pred_stats(_y, _y_pred)
                    text = f"R2={st.r2:.3f} MAE={st.mae:.3f}"
                    x_min = _x[_x.argmin()]
                    x_max = _x[_x.argmax()]
                    y_min = _y_pred[_x.argmin()]
                    y_max = _y_pred[_x.argmax()]

                    plt.text(.01, .99, text, ha='left', va='top', transform=ax.transAxes)
                    plt.plot([x_min, x_max], [y_min, y_max], ls=':', color='black')

            elif kind == 'line':
                if len(df_g) == 1:
                    params['marker'] = 'x'

                df_g = df_g.sort_values([x, y])
                plt.plot(df_g[x], df_g[y], **params)

            elif kind == 'hdi':
                hdi_probs = sorted(hdi_probs)
                alpha_step = alpha / len(hdi_probs)
                g_hdi = df_g.groupby(x)
                for hdi_prob in hdi_probs:
                    try:
                        df_hdi = g_hdi.apply(lambda dfx: pd.Series(az.hdi(dfx[y].to_numpy(), hdi_prob=hdi_prob), index=['low', 'high']))
                    except ValueError:
                        print('hi')
                        raise
                    df_hdi = df_hdi.reset_index()
                    params['label'] = label
                    params['alpha'] = alpha_step
                    if 'color' not in params:
                        params['color'] = 'blue'

                    plt.fill_between(df_hdi[x], df_hdi['low'], df_hdi['high'], linewidth=0, **params)
                    label = None

            else:
                raise ValueError(kind)

        if title and plot_title:
            plot_title = f"{title}: {plot_title}"

        elif title:
            plot_title = title

        if plot_title:
            plt.title(plot_title)

        if labels and legend_loc != 'off':
            plt.legend(loc=legend_loc)

        if x_axis_type == int:
            plt.gca().xaxis.set_major_locator(mticker.MultipleLocator(1))

        if y_axis_type == int:
            plt.gca().yaxis.set_major_locator(mticker.MultipleLocator(1))

        if xlim:
            plt.xlim(xlim)

        if ylim:
            plt.ylim(ylim)

        if x:
            plt.xlabel(xsettings.x_get_desc(x))

        if y:
            plt.ylabel(xsettings.x_get_desc(y))

        if add_date:
            date_str = dt.date.today().isoformat()
            ax.annotate(date_str, xy=(0.9, -0.08), xycoords='axes fraction')

        plt.tight_layout()

        if save_to:
            plot_title = plot_title or "all"
            plt.savefig(save_to.joinpath(f"{plot_title}.png"), pad_inches=0)

        elif show:
            plt.show()



import unittest
import matplotlib.pyplot as plt
from ..xpd import *
from .. import xplt

monkey_patch()
xplt.monkey_patch()


class TestXPD(unittest.TestCase):
    def test_decorate(self):
        df = pandas.DataFrame({'Aaaaa': [1, 1, 2, 2, 3, 3], 'Bbbbb': [1, 2, 3, 4, 5, 6]})
        df.x_rename(columns={'Aaaaa': 'a', 'Bbbbb': 'b'}, inplace=True)
        df.x_rename(columns={'a': 'my_a'}, inplace=True)
        plt.scatter(df.my_a, df.b); plt.show()
        plt.hist(df.my_a); plt.show()



import tempfile
from collections import Counter
import matplotlib.pyplot as plt
from slugify import slugify
from PIL import Image
from scriptine import path
import pptx
from pptx.util import Inches, Cm
from xdat import xsettings


DEFAULT_THEME = xsettings.XDAT_ROOT.joinpath('media', 'default_theme.pptx')
assert DEFAULT_THEME.exists(), DEFAULT_THEME


class Presentation:
    def __init__(self, theme=DEFAULT_THEME):
        self.theme = theme
        self.prs = pptx.Presentation(self.theme)

    def add_slide(self, layout_name, **kwargs):
        add_slide(self.prs, layout_name, **kwargs)

    def print_layout(self):
        print_layout(self.theme)

    def save(self, out_path):
        self.prs.save(out_path)

    @classmethod
    def capture_image(cls):
        return Img()


class Img:
    """
    Everything's in inches...
    """
    DPI = 80

    def __init__(self, tight_layout=True):
        assert xsettings.CACHE_PATH is not None, "must set xsettings.CACHE_PATH"
        tmp_folder = xsettings.CACHE_PATH.joinpath('xpptx')
        tmp_folder.ensure_dir()

        self.img_path = tempfile.NamedTemporaryFile(suffix='.png', delete=False, dir=tmp_folder).name

        if tight_layout:
            plt.tight_layout()

        plt.savefig(self.img_path, pad_inches=0)
        plt.close('all')
        img = Image.open(self.img_path)
        w, h = img.size
        self.width = w/self.DPI
        self.height = h/self.DPI

    def box(self, width, height):
        rw = self.width / width
        rh = self.height / height
        rmax = max(rh, rw)
        return self.width/rmax, self.height/rmax

    def __str__(self):
        return self.img_path

    def __repr__(self):
        return self.img_path

    def __lt__(self, other):
        return False


def _slug(text):
    return slugify(text, separator="_")


def _slug_dict(input_dict):
    counts = Counter()
    out_dict = dict()
    for text, v in input_dict.items():
        try:
            parts = text.split()
            int(parts[-1])
            parts = parts[:-1]
            text = " ".join(parts)
        except:
            pass

        try:
            parts = text.split()
            if parts[-1].lower() == 'placeholder':
                parts = parts[:-1]
            text = " ".join(parts)
        except:
            pass

        text = _slug(text)
        counts[text] += 1

        if counts[text] > 1:
            text = f"{text}_{counts[text]}"

        out_dict[text] = v

    return out_dict


def _get_layout(prs):
    layouts = _slug_dict({l.name: l for l in prs.slide_layouts})
    return layouts


def _get_placeholders(slide):
    placeholders = _slug_dict({p.name: p for p in slide.placeholders})
    return placeholders


def add_slide(prs, layout_name, **kwargs):
    layouts = _get_layout(prs)
    assert layout_name in layouts, f"{layout_name} not in: {sorted(layouts)}"

    layout = layouts[layout_name]
    slide = prs.slides.add_slide(layout)

    placeholders = _get_placeholders(slide)

    for k, v in kwargs.items():
        assert k in placeholders, f"{k} not in: {sorted(placeholders)}"
        p = placeholders[k]

        if isinstance(v, str):
            p.text = v

        elif isinstance(v, Img):
            w,h = v.box(p.width, p.height)
            w = int(w)
            h = int(h)
            slide.shapes.add_picture(str(v), p.left, p.top, height=h)
            p.text = ' '

        else:
            raise TypeError(type(v))

    for k in set(placeholders) - set(kwargs):
        p = placeholders[k]
        p.text = ' '

    return


def print_layout(template_path=None):
    prs = pptx.Presentation(template_path)
    layouts = _get_layout(prs)
    for layout_name in sorted(layouts):
        print(f"- {layout_name}")
        layout = layouts[layout_name]
        slide = prs.slides.add_slide(layout)
        placeholders = _get_placeholders(slide)
        for pl in sorted(placeholders):
            print(f"  + {pl}")

    return


if __name__ == "__main__":
    xsettings.PROJECT_NAME = 'xdat'
    xsettings.updated_config()
    plt.scatter([1, 2], [3, 4])
    i = Img()
    w, h = i.box(5, 5)

    prs = Presentation()
    prs.print_layout()

    prs.add_slide('title_and_body', title='hi', text=i)
    prs.save('/tmp/xdat/test.pptx')



import inspect
import warnings
import functools
import hashlib
import cloudpickle as pickle
import pandas as pd

from . import xsettings
from scriptine import path


def x_cached(name='', hash_key=None, also_parquet=False, outer_level=1, static=False):
    def decorator(func):
        curframe = inspect.currentframe()
        calframe = inspect.getouterframes(curframe, 2)
        calling_name = calframe[outer_level][3]
        if calling_name == '<module>':
            calling_name = path(calframe[outer_level][1]).namebase

        f_src = inspect.getsource(func)
        short_name = name or func.__name__
        f_name = f"{calling_name}__{func.__name__}"
        if name:
            f_name = f"{f_name}__{name}"

        @functools.wraps(func)
        def _cached(*args, **kwargs):
            actually_static = static
            cache_folder = None

            if actually_static:
                if not xsettings.STATIC_CATH_PATH:
                    warnings.warn('xsettings.STATIC_CATH_PATH is not set: setting static=False')
                    actually_static = False
                else:
                    cache_folder = xsettings.STATIC_CATH_PATH

            if not actually_static:
                assert xsettings.CACHE_PATH is not None, "must set xsettings.CACHE_PATH"
                cache_folder = xsettings.CACHE_PATH

            assert cache_folder, "Problem with cache_folder"

            cache_folder = path(cache_folder)
            cache_folder = cache_folder.joinpath(f_name)

            code_text = f_src + f_name + name + str(hash_key) + str((args, kwargs))
            code_hash = hashlib.md5(code_text.encode('utf-8')).hexdigest()

            cache_subfolder = cache_folder.joinpath(code_hash)
            cache_subfolder.ensure_dir()
            cache_file = cache_subfolder.joinpath(f"{short_name}.pickle")

            if cache_file.exists():
                with open(cache_file, 'rb') as f:
                    return pickle.load(f)

            results = func(*args, **kwargs)

            with open(cache_file, 'wb') as f:
                pickle.dump(results, f)

            if also_parquet:
                assert isinstance(results, pd.DataFrame), "need DataFrame to save as parquet"
                parquet_path = cache_subfolder.joinpath(f"{short_name}.parquet")
                results.to_parquet(parquet_path, use_deprecated_int96_timestamps=True)

            return results
        return _cached

    return decorator


def x_cached_call(func, *args, name='', hash_key=None, also_parquet=False, static=False, cached=True, **kwargs):
    if cached:
        dec = x_cached(name=name, hash_key=hash_key, also_parquet=also_parquet, outer_level=2, static=static)
        return dec(func)(*args, **kwargs)

    return func(*args, **kwargs)



import os
import tempfile

import matplotlib.colors as mcolors
from scriptine import path

PROJECT_NAME = 'xdat'
IGNORE_INDEX = True                 # useful when dataframes don't have a meaningful index
SANITY_CHECKS = False               # False, "warn", "error"
NAN_TEXTS = ['', '-', '.', '--', 'X', 'x', '#DIV/0!', '#VALUE!', 'NA', 'NAN', '0R', 'NR', 'NULL', 'None']
ACCU_METHOD = 'fsum'                # method to perform accurate math

FIGSIZE = (10, 10)

COL_DESC = dict()                   # column name --> column description
DEFAULT_COLORS = mcolors.TABLEAU_COLORS.keys()
HARD_CODED_COLORS = {
    '0': 'xkcd:sea blue',
    '0.0': 'xkcd:sea blue',
    '1': 'xkcd:crimson',
    '1.0': 'xkcd:crimson',
    'NaN': 'xkcd:gray',
    'None': 'xkcd:gray',
}

DEFAULT_MARKERS = 'o^sP*DXp<>v'
HARD_CODED_MARKERS = {
    '0': 'o',
    '0.0': 'o',
    '1': 'X',
    '1.0': 'X',
    'NaN': '^',
    'None': '^',
}

OUTPUT_PATH = path()
CACHE_PATH = path()
STATIC_CATH_PATH = None
XDAT_ROOT = path(__file__).parent.abspath()


def x_add_desc(from_name, to_name):
    if from_name == to_name:
        return

    COL_DESC[from_name] = to_name


def x_get_desc(key):
    for _ in range(100):
        if key not in COL_DESC:
            return key

        assert COL_DESC[key] != key
        key = COL_DESC[key]

    raise ValueError(f"Looks like some sort of endless loop for '{key}'")


def updated_config(project_name=None, verbose=True, static_cache_folder=None):
    global OUTPUT_PATH, CACHE_PATH, PROJECT_NAME, STATIC_CATH_PATH

    PROJECT_NAME = project_name or PROJECT_NAME
    assert PROJECT_NAME, "need to set PROJECT_NAME"

    base_path = os.environ.get('DSXP_OUTPUT_PATH', tempfile.gettempdir())

    OUTPUT_PATH = path(base_path).joinpath(PROJECT_NAME)
    CACHE_PATH = OUTPUT_PATH.joinpath('cache')
    OUTPUT_PATH.ensure_dir()

    if static_cache_folder:
        STATIC_CATH_PATH = path(static_cache_folder)
        STATIC_CATH_PATH.ensure_dir()

    if verbose:
        print(f"OUTPUT_PATH set to '{OUTPUT_PATH}'")

    return OUTPUT_PATH


def get_default(default, override, possible_values=None):
    if override is None:
        if possible_values:
            assert default in possible_values, default

        return default

    if possible_values:
        assert override in possible_values, override

    return override


updated_config(verbose=False)



import unittest
import numpy as np
from ..xstats import *
from ..xchecks import *


class TestXStats(unittest.TestCase):
    def test_x_model_pred_stats(self):
        y_true = np.ones(5)
        y_pred = np.linspace(0.75, .95, len(y_true))
        y_alt = 0.8
        s1 = x_model_pred_stats(y_true, y_pred, y_alt)
        self.assertIsNone(s1.p_value_err_normal)

        y_true = np.ones(20)
        y_pred = np.linspace(0.75, .95, len(y_true))
        y_alt = 0.8
        s2 = x_model_pred_stats(y_true, y_pred, y_alt, k=1)
        s3 = x_model_pred_stats(y_true, y_pred, y_alt, k=5)

        self.assertGreater(s1.p_value, s2.p_value)
        self.assertGreater(s2.r2_adj, s3.r2_adj)
        self.assertIsNotNone(s2.p_value_err_normal)



from typing import Optional, Dict, Any
import makefun
import inspect
from sklearn.model_selection import cross_val_predict
from baikal import Input, Model, Step
from baikal import make_step as _make_step_orig
from baikal.steps import Lambda, Split, Stack, ColumnStack, Concatenate
from baikal.plot import plot_model

"""
See: https://baikal.readthedocs.io/en/stable/index.html
"""


def make_step_wrapper(base_class: type, attr_dict: Dict[str, Any] = None, class_name: Optional[str] = None) -> type:
    if class_name is None and hasattr(base_class, '__name__'):
        class_name = base_class.__name__

    step_orig = _make_step_orig(base_class, attr_dict=attr_dict, class_name=class_name)
    sig = inspect.signature(base_class)
    step_fixed = makefun.with_signature(sig)(step_orig)
    return step_fixed


assert inspect.signature(_make_step_orig) == inspect.signature(make_step_wrapper)


def make_step(base_class: type, attr_dict: Dict[str, Any] = None, class_name: Optional[str] = None, add_fit_predict=False, cv_pred=None) -> type:
    if add_fit_predict:
        attr_dict = attr_dict if attr_dict else dict()
        if cv_pred is None:
            cv_pred = lambda self, X, y: cross_val_predict(self, X, y, method="predict_proba")

        def fit_predict(self, X, y):
            self.fit(X, y)
            return cv_pred(self, X, y)

        attr_dict['fit_predict'] = fit_predict

    return make_step_wrapper(base_class, attr_dict=attr_dict, class_name=class_name)


if __name__ == "__main__":
    from sklearn import svm

    SVC = make_step(svm.SVC, add_fit_predict=True)
    clf = SVC(C=0.1)
    print('bye')



import pandas as pd
import numpy as np
import datetime as dt
from tqdm import tqdm
import matplotlib.pyplot as plt
import seaborn as sns
import missingno as msno
import feature_engine
from scriptine import path
from caseconverter import snakecase as snake_case
from munch import Munch as MunchDict
from joblib import Parallel, delayed

from . import xsettings, xagg, xcache, xchecks, xmunge, xnp, xpd, xplt, xutils, xdata, xproblem, xplots, xstats, xstan, xpptx, xfactors, x1d, xeda, xcalc, xparallel, xmodels
from .xutils import x_monkey_patch

from .xsettings import x_add_desc, x_get_desc
from .xcache import x_cached, x_cached_call
from .xplots import post_plot



import unittest
from ..xpd import *
from ..xchecks import *

monkey_patch()


class TestXPD(unittest.TestCase):
    def test_x_groupby(self):
        df = pandas.DataFrame({'a': [1, 1, 2, 2, 3, 3], 'b': [1, 2, 3, 4, 5, 6]})
        df_g = df.x_groupby(['a'], {'max_b': xagg.Max('b'), 'any_b': xagg.Any('b'), 'xx_b': xagg.Lambda('b', np.mean), 'n': xagg.Count()})
        check_same_sets(df['a'], df_g['a'])
        check_no_dups(df_g['a'])

    def test_x_calc_rank_num(self):
        df = pandas.DataFrame({'a': [1, 1, 2, 2, 3, 3], 'b': [1, 2, 3, 4, 5, 6]})
        df['rank_id'] = df.x_calc_rank_num('a', 'b')
        check_same_sets([1, 2], df.rank_id)

    def test_x_history(self):
        df = pandas.DataFrame({'a': [1, 1, 2, 2, 3, 3], 'b': [1, 2, 3, 4, 5, 6]})
        df.x_add_history('h1', 'yo')
        return



from xdat.xincludes import *
from sklearn.ensemble import RandomForestClassifier
from sklearn import model_selection
from lightgbm.sklearn import LGBMClassifier


def main():
    df = x_cached_call(xdata.read_csv, "/data/kaggle/loan_eligibility/Loan_Data.csv")
    xeda.x_inspect_cols(df)
    df['loan_id'] = df.loan_id.apply(lambda t: int(t[2:]))
    df['dependents'] = xpd.x_replace(df.dependents, replace_vals={'0': 0, '1': 1, '2': 2, '3+': 4})
    df['total_income'] = df.applicant_income + df.coapplicant_income
    df['applicant_ratio'] = df.loan_amount / df.applicant_income
    df['total_ratio'] = df.loan_amount / df.total_income
    print('-----')
    xeda.x_inspect_cols(df)

    # binary_classification(df)
    ordinal_classifier(df)


def ordinal_classifier(df):
    clf = RandomForestClassifier(class_weight='balanced', min_samples_leaf=0.03, max_depth=5)
    oc = xmodels.OrdinalClassifier(clf)

    df, new_target = xeda.x_prep_df(df, 'dependents')
    X, y = xutils.split_X_y(df, new_target)
    X_train, X_test, y_train, y_test = model_selection.train_test_split(X, y, train_size=0.7, random_state=42)

    oc.fit(X_train, y_train)
    print('test score:', oc.score(X_test, y_test))
    return


def binary_classification(df):
    # df['property_area_ord'] = xpd.x_replace(df.property_area, replace_vals={'Rural': 0, 'Semiurban': 0.5, 'Urban': 1.0})
    df_corr = xeda.x_corr_with_target(df, 'loan_status')
    cols_drop = set(df_corr[df_corr.abs_corr < 0.03].orig_col.values)
    cols_drop.add('loan_id')
    df.drop(columns=list(cols_drop), inplace=True)

    df, new_target = xeda.x_prep_df(df, 'loan_status')
    # X, y = xutils.split_X_y(df, new_target)
    # df_2d = xeda.x_reduce_dim(X, method='umap')
    # df_2d[new_target] = y
    # xplots.plot_multi(df_2d, x='x', y='y', color_on=new_target, figsize=(6,6))

    df.rename(columns={new_target: 'target'}, inplace=True)

    pipe = RandomForestClassifier(class_weight='balanced', min_samples_leaf=0.03, max_depth=5)
    pipe = LGBMClassifier(class_weight='balanced', max_depth=10, num_leaves=20)
    pipe = LGBMClassifier(class_weight='balanced', max_depth=10, boosting_type='rf', bagging_freq=2, bagging_fraction=0.8, n_estimators=500)

    df_test, all_folds = xproblem.train_cv(df, 'target', pipe, stratify_on='target', eval_size=0.2, eval_type='lightgbm', fit_params={'early_stopping_rounds': 500, 'eval_metric': 'auc'})

    xplots.plot_roc_curve(all_folds[0].df_train.target, all_folds[0].df_train.prob_1, title='TRAIN')

    xplots.plot_roc_curve(all_folds[0].df_val.target, all_folds[0].df_val.prob_1, title='VALIDATION')

    xplots.plot_confusion_matrix(df_test.target, df_test.pred, y_score=df_test.prob_1)

    df_eval = xproblem.eval_test(df_test, eval_per=['none'], metric_list=['AUC'])
    print(df_eval.to_string(index=False))

    print('Corr between model score & target:', np.corrcoef(df_test.prob_1, df_test.target)[0][1])

    xplots.plot_roc_curve(df_test.target, df_test.prob_1)
    xplots.plot_model_scores(df_test.target, df_test.prob_1)
    return


if __name__ == "__main__":
    main()



import numpy as np
import pandas as pd
from collections import namedtuple
from sklearn import model_selection, pipeline, metrics
from sklearn.base import clone
from tqdm import tqdm
from joblib import Parallel, delayed
from . import xutils


Metric = namedtuple('Metric', ['name', 'func', 'args'])
METRICS = {
    'AUC': Metric(name='AUC', func=metrics.roc_auc_score, args=['target', 'prob_1']),
    'R2': Metric(name='R2', func=metrics.r2_score, args=['pred', 'target']),
    'MAPE': Metric(name='MAPE', func=metrics.mean_absolute_percentage_error, args=['pred', 'target']),
    'MAE': Metric(name='MAE', func=metrics.mean_absolute_error, args=['pred', 'target']),
    'MSE': Metric(name='MSE', func=metrics.mean_squared_error, args=['pred', 'target']),
    'MSLE': Metric(name='MSLE', func=metrics.mean_squared_log_error, args=['pred', 'target']),
}

for k,v in list(METRICS.items()):
    METRICS[k.lower()] = v


CVSplit = namedtuple("CVSplit", ['df_train', 'df_test', 'df_val'])


def cv_split(df, n_splits=12, stratify_on=None, group_on=None, val_size=0.0):
    df = df.sample(frac=1.).reset_index(drop=True)
    groups = df[group_on] if group_on else None
    y = df[stratify_on] if stratify_on else None

    if group_on is None:
        if n_splits == 'max':
            n_splits = len(df)

        if stratify_on:
            kfold = model_selection.StratifiedKFold(n_splits=n_splits)

        else:
            kfold = model_selection.KFold(n_splits=n_splits)

    else:
        num_groups = len(np.unique(groups))
        if n_splits == 'max':
            n_splits = num_groups

        n_splits = min(n_splits, num_groups)

        if stratify_on:
            kfold = model_selection.StratifiedGroupKFold(n_splits=n_splits)

        else:
            kfold = model_selection.GroupKFold(n_splits=n_splits)

    for train_index, test_index in kfold.split(df, y=y, groups=groups):
        df_train = df.iloc[train_index].reset_index(drop=True)
        df_test = df.iloc[test_index].reset_index(drop=True)
        df_val = None

        if val_size:
            assert not group_on, "Not yet implemented"

            stratify = df_train[stratify_on] if stratify_on else None
            df_train, df_val = model_selection.train_test_split(df_train, test_size=val_size, stratify=stratify)

        yield CVSplit(df_train=df_train, df_test=df_test, df_val=df_val)


CVFold = namedtuple("CVFold", ['clf', 'df_train', 'df_test', 'feature_names', 'df_val'])


def train_cv(df, target_col, clf, n_splits=12, stratify_on=None, group_on=None, ordered_split=False, del_cols=tuple(), uid_col=None, sample_weight_col=None, eval_size=0.0, eval_type=None, fit_params=None, n_jobs=1):
    if eval_size:
        assert eval_type in ['lightgbm'], "need to specify proper eval_type param"
    else:
        eval_type = None

    fit_params = dict() if fit_params is None else fit_params

    del_cols = set(del_cols)
    if uid_col:
        del_cols.add(uid_col)
    if sample_weight_col:
        del_cols.add(sample_weight_col)
    del_cols = sorted(del_cols)

    def calc_fold(fold_num, split_data: CVSplit):
        clf_fold = clone(clf)
        df_train = split_data.df_train
        df_train_data = df_train.copy()
        df_val = split_data.df_val
        df_val_data = None
        if df_val is not None:
            df_val_data = df_val.copy()

        df_test = split_data.df_test
        df_test_data = df_test.copy()
        test_tmp_uid = df_test[uid_col] if uid_col else None
        sample_weight_train = df_train[sample_weight_col] if sample_weight_col else None
        sample_weight_eval = df_val[sample_weight_col] if sample_weight_col else None

        if ordered_split and group_on:
            assert n_splits == 'max'
            test_group = df_test[group_on].min()
            df_train = df_train[df_train[group_on] < test_group].reset_index(drop=True)
            assert (df_train[group_on] >= test_group).sum() == 0
            if len(df_train) == 0:
                return

        if del_cols:
            df_train_data = df_train.drop(columns=del_cols, errors='ignore')
            df_test_data = df_test.drop(columns=del_cols, errors='ignore')
            if df_val_data is not None:
                df_val_data = df_val_data.drop(columns=del_cols, errors='ignore')

        X_train, y_train = xutils.split_X_y(df_train_data, target_col)
        X_eval, Y_eval = None, None
        if df_val_data is not None:
            X_eval, Y_eval = xutils.split_X_y(df_val_data, target_col)

        if sample_weight_train is not None:
            if eval_type == 'lightgbm':
                eval_set = [[X_eval, Y_eval]]
                clf_fold.fit(X_train, y_train, sample_weight=sample_weight_train, eval_set=eval_set, eval_sample_weight=sample_weight_eval, **fit_params)
            else:
                clf_fold.fit(X_train, y_train, sample_weight=sample_weight_train, **fit_params)

        else:
            if eval_type == 'lightgbm':
                eval_set = [[X_eval, Y_eval]]
                clf_fold.fit(X_train, y_train, eval_set=eval_set, **fit_params)
            else:
                clf_fold.fit(X_train, y_train, **fit_params)

        pred_train = clf_fold.predict(X_train)
        df_train['pred'] = pred_train
        if hasattr(clf_fold, "predict_proba"):
            prob_1 = clf_fold.predict_proba(X_train)[:, 1]
            df_train['prob_1'] = prob_1

        if eval_type:
            pred_eval = clf_fold.predict(X_eval)
            df_val['pred'] = pred_eval
            if hasattr(clf_fold, "predict_proba"):
                prob_1 = clf_fold.predict_proba(X_eval)[:, 1]
                df_val['prob_1'] = prob_1

        X_test, y_test = xutils.split_X_y(df_test_data, target_col)
        pred = clf_fold.predict(X_test)
        df_test['pred'] = pred

        if hasattr(clf_fold, "predict_proba"):
            prob_1 = clf_fold.predict_proba(X_test)[:, 1]
            df_test['prob_1'] = prob_1

        df_test['fold_num'] = fold_num
        if uid_col:
            df_test[uid_col] = test_tmp_uid

        fold = CVFold(clf=clf_fold, df_train=df_train, df_test=df_test, feature_names=X_train.columns, df_val=df_val)

        return fold

    splits = cv_split(df, n_splits=n_splits, stratify_on=stratify_on, group_on=group_on, val_size=eval_size)
    total = len(df) if n_splits == 'max' else n_splits
    tqdm_splits = tqdm(enumerate(splits), total=total)

    if n_jobs == 1:
        all_folds = [calc_fold(fold_num, split_data) for fold_num, split_data in tqdm_splits]

    else:
        all_folds = Parallel(n_jobs=n_jobs)(delayed(calc_fold)(fold_num, split_data) for fold_num, split_data in tqdm_splits)

    all_folds = [fold for fold in all_folds if fold is not None]

    df_test = pd.concat([i[2] for i in all_folds], ignore_index=True)
    df_test = df_test.reset_index(drop=True)
    return df_test, all_folds


def eval_test(df_test, eval_per=None, metric_list=None):
    """

    :type metrics: list[Metric|str]
    """

    df_test = df_test.copy()
    df_test['none'] = 'none'

    metric_list = [m if isinstance(m, Metric) else METRICS[m] for m in metric_list]

    if not isinstance(eval_per, list) and not isinstance(eval_per, tuple):
        eval_per = [eval_per]

    rows = []
    for curr_group in eval_per:
        if not curr_group:
            curr_group = 'none'

        for gval in df_test[curr_group].unique():
            df_g = df_test[df_test[curr_group] == gval]
            row = {m.name: m.func(*(df_g[arg] for arg in m.args)) for m in metric_list}
            row['cv_grouping'] = curr_group
            row['cv_group_key'] = str(gval)
            row['cv_group_size'] = len(df_g)
            rows.append(row)

    df_res = pd.DataFrame(rows)
    return df_res


def calc_feature_importances(folds, flat=False):
    rows = []
    fi_var = None
    for fold in folds:
        clf = fold.clf
        if isinstance(clf, pipeline.Pipeline):
            clf = clf[-1]

        if fi_var is None:
            for var in ['feature_importances_', 'coef_']:
                if hasattr(clf, var):
                    fi_var = var
                    break

            if not fi_var:
                print("Can't find in clf")
                return

        fi = getattr(clf, fi_var).flatten()
        fi_dict = dict(zip(fold.feature_names, fi))
        if flat:
            for k, v in fi_dict.items():
                rows.append({'feature_name': k, 'feature_importance': v})

        else:
            rows.append(fi_dict)

    df = pd.DataFrame(rows)
    return df


class ModelEnsemble:
    def __init__(self, clfs):
        self.clfs = clfs

    @classmethod
    def from_all_folds(cls, all_folds):
        clfs = [f.clf for f in all_folds]
        return cls(clfs)

    def predict(self, X):
        all_y = [clf.predict(X) for clf in self.clfs]
        all_y = [y.reshape(y.shape + (1,)) for y in all_y]
        Y = np.concatenate(all_y, axis=1)
        pred = [np.bincount(y).argmax() for y in Y]
        pred = np.array(pred)
        return pred

    def predict_binary(self, X, threshold=0.5):
        prob_1 = self.predict_proba(X)[:, 1]
        pred = prob_1 >= threshold
        return pred.astype(int)

    def predict_proba(self, X):
        all_y = [clf.predict_proba(X) for clf in self.clfs]
        all_y = [y.reshape(y.shape + (1,)) for y in all_y]
        y = np.concatenate(all_y, axis=2)
        proba = y.mean(axis=2)
        return proba




import pandas as pd
import matplotlib.pyplot as plt
import seaborn._core
import seaborn.relational

from . import xsettings


def set_long_column_names(df):
    df = df.copy()

    renames = dict()
    for col in df.columns:
        new_name = xsettings.x_get_desc(col)
        renames[col] = new_name

    df.rename(columns=renames, inplace=True)
    return df


def unslug(text):
    """
    Given slugified text, return unsligified version
    """
    parts = text.split('_')
    parts = [p.capitalize() for p in parts]
    new_text = " ".join(parts)
    return new_text


def match_key_value(values, hard_coded=None):
    hard_coded = hard_coded or dict()
    seen = dict()

    def get_value(key):
        key = str(key)

        if key in hard_coded:
            return hard_coded[key]

        if key in seen:
            return seen[key]

        seen[key] = values[len(seen) % len(values)]
        return seen[key]

    get_value.seen = seen
    get_value.values = values
    get_value.hard_coded = hard_coded

    return get_value


def match_colors(default_colors=None, hard_coded_colors=None):
    default_colors = xsettings.get_default(xsettings.DEFAULT_COLORS, default_colors)
    hard_coded_colors = xsettings.get_default(xsettings.HARD_CODED_COLORS, hard_coded_colors)

    return match_key_value(default_colors, hard_coded=hard_coded_colors)


def match_markers(default_markers=None, hard_coded_markers=None):
    default_markers = xsettings.get_default(xsettings.DEFAULT_MARKERS, default_markers)
    hard_coded_markers = xsettings.get_default(xsettings.HARD_CODED_MARKERS, hard_coded_markers)

    return match_key_value(default_markers, hard_coded=hard_coded_markers)


def update_axes(x=None, y=None):
    if isinstance(x, pd.Series):
        x = x.name

    if isinstance(y, pd.Series):
        y = y.name

    try:
        x = xsettings.x_get_desc(x)
        y = xsettings.x_get_desc(y)
    except TypeError:
        return

    if x:
        plt.xlabel(x)

    if y:
        plt.ylabel(y)


def decorate(x=None, y=None, xlim=None, ylim=None, title=None, show=False):
    update_axes(x=x, y=y)

    if xlim:
        plt.xlim(xlim)

    if ylim:
        plt.ylim(ylim)

    if title:
        plt.title(title)

    # plt.legend()
    plt.tight_layout()

    if show:
        plt.show()


def _sns_vp__init__(self, *args, **kwargs):
    """
    Fails for categorical that are actually numeric... need to fix
    """
    if 'data' in kwargs and isinstance(kwargs['data'], pd.DataFrame) and kwargs.get('variables', dict()).get('hue'):
        kwargs['data'] = kwargs['data'].sort_values(kwargs['variables']['hue']).reset_index(drop=True)

    return self._orig__init__(*args, **kwargs)


def _sns_add_axis_labels(self, ax, default_x="", default_y=""):
    """Add axis labels if not present, set visibility to match ticklabels."""
    # TODO ax could default to None and use attached axes if present
    # but what to do about the case of facets? Currently using FacetGrid's
    # set_axis_labels method, which doesn't add labels to the interior even
    # when the axes are not shared. Maybe that makes sense?
    if not ax.get_xlabel():
        x_visible = any(t.get_visible() for t in ax.get_xticklabels())
        label = self.variables.get("x", default_x)
        label = xsettings.x_get_desc(label)
        ax.set_xlabel(label, visible=x_visible)
    if not ax.get_ylabel():
        y_visible = any(t.get_visible() for t in ax.get_yticklabels())
        label = self.variables.get("y", default_y)
        label = xsettings.x_get_desc(label)
        ax.set_ylabel(label, visible=y_visible)


def _sns_rp_add_legend_data(self, *args, **kwargs):
    res = self._orig_add_legend_data(*args, **kwargs)
    self.legend_title = xsettings.x_get_desc(self.legend_title)
    return res


def monkey_patch_sns():
    # seaborn._core.VectorPlotter._add_axis_labels = _sns_add_axis_labels

    # seaborn._core.VectorPlotter._orig__init__ = seaborn._core.VectorPlotter.__init__
    # seaborn._core.VectorPlotter.__init__ = _sns_vp__init__

    seaborn.relational._RelationalPlotter._orig_add_legend_data = seaborn.relational._RelationalPlotter.add_legend_data
    seaborn.relational._RelationalPlotter.add_legend_data = _sns_rp_add_legend_data


def monkey_patch_x_y(package, fname):
    f = getattr(package, fname)

    def f_new(x, y, *args, **kwargs):
        ret = f(x, y, *args, **kwargs)
        update_axes(x, y)
        return ret

    setattr(package, f"_{fname}_orig", f)
    setattr(package, fname, f_new)


def monkey_patch_x(package, fname):
    f = getattr(package, fname)

    def f_new(x, *args, **kwargs):
        ret = f(x, *args, **kwargs)
        update_axes(x)
        return ret

    setattr(package, f"_{fname}_orig", f)
    setattr(package, fname, f_new)


def monkey_patch():
    for fname in ['scatter', 'plot']:
        monkey_patch_x_y(plt, fname)

    for fname in ['hist']:
        monkey_patch_x(plt, fname)

    monkey_patch_sns()




import pandas as pd
from tqdm import tqdm
from joblib import Parallel, delayed


def array_col_to_wide(df, array_col, prefix=None, n_jobs=1):
    if prefix is None:
        prefix = f"{array_col}"

    def get_row(row):
        a = row[array_col]
        del row[array_col]
        for idx, v in enumerate(a):
            row[f"{prefix}_{idx}"] = v
        return row

    all_rows = Parallel(n_jobs=n_jobs)(delayed(get_row)(r[1]) for r in tqdm(df.iterrows(), total=len(df)))
    df_all = pd.DataFrame(all_rows)
    return df_all



X_TARGET = 'target'     # a a target variable
X_META = 'meta'         # meta data (not to use as training feature)
X_KEY = 'key'           # a sample key (should be unique)
X_FEATURE = 'feature'   # a feature (can be used for training)

X_BINARY = 'binary'                # a binary feature
X_CATEGORICAL = 'categorical'      # a categorical feature
X_ORDINAL = 'ordinal'
X_INTERVAL = 'interval'
X_RATIO = 'ratio'




from setuptools import setup
import pathlib

"""
Can install dependencies directly by running from within this folder:
> pip install .
"""

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name='xdat',
    version='0.1.70',
    description='eXtended Data Analysis Toolkit',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://bitbucket.org/hermetric/xdat/',
    author='Ido Carmi',
    author_email='ido@hermetric.com',
    license='MIT',
    packages=['xdat'],
    install_requires=['pandas',
                      'numpy',
                      'scikit-learn',
                      'scriptinep3',
                      # 'pystan<3',
                      'tqdm',
                      'joblib',
                      'cloudpickle',
                      'matplotlib',
                      'pandas-sets',
                      'python-slugify',
                      # 'accupy',  # requires > sudo apt -q -y install libeigen3-dev
                      # 'tensorflow>=2',
                      'seaborn',
                      'missingno',
                      'data-science-utils',
                      'munch',
                      'arviz',      # vs inference-tools
                      'python-pptx',
                      'feature_engine',
                      'case-converter',
                      'datashader',
                      'umap-learn',
                      'networkx',
                      'lightgbm',
                      'pydot',
                      'baikal',
                      'makefun'
                      ],

    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
)



import matplotlib.pyplot as plt
import pandas as pd
import scipy.stats as stats
from xpd import *

import xplt

monkey_patch()
xplt.monkey_patch()


def xplot_scatter(data, feature, target, per_categ=None):
    fig, ax = plt.subplots(figsize=(10, 6))

    if per_categ:
        subsets = data[per_categ].unique()
        for s in subsets:
            temp = data[data[per_categ]==s]
            ax.scatter( temp[feature], temp[target], label = s)
    else:
        ax.scatter(data[feature], data[target])
    xplt.update_axes(feature, target)
    xplt.decorate(title= feature + ' vs. ' + target, show=True)


def xplot_feature_distribution(data, feature, target):

    fig, ax = plt.subplots(figsize=(10, 6))
    labels = data[target].unique()
    for l in labels:
        values = data[data[target]==l][feature]
        values = values.sort_values()
        mean = np.mean(values)
        std = np.std(values)
        pdf = stats.norm.pdf(values, mean, std)
        ax.plot(values, pdf, label=l)

    xplt.update_axes(target, target + ' pdf')
    xplt.decorate(title='Distribution of '+feature + ' per ' + target + ' values', show=True)


def xsee_data(data, target, target_type='class', show='all'):

    features = list(data.columns)
    features.remove(target)
    if target_type == 'class':
        for feat in features:
            if data[feat].dtype.kind in 'iufc':
                xplot_feature_distribution(data, feat, target)

    elif target_type =='continuous':
        for feat in features:
            if data[feat].dtype.kind in 'iufc':
                xplot_scatter(data, feat, target)
            else:
                xplot_feature_distribution(data, target, feat)



def runtoy():

    data = pd.read_csv('../toy_data/student-mat.csv')
    print(data.columns)
    xsee_data(data, 'G3', 'continuous')

if __name__ == "__main__":
    runtoy()


import numpy
try:
    import accupy
except (ImportError, ModuleNotFoundError):
    pass

from . import xsettings


def x_sum(a, accu_method=None):
    """
    A much slower, but more accurate sum routine
    (see https://github.com/nschloe/accupy)
    """

    accu_method = xsettings.get_default(xsettings.ACCU_METHOD, accu_method)

    if accu_method == 'fsum':
        aa = accupy.fsum(a)
    elif accu_method == 'ksum':
        aa = accupy.ksum(a)
    elif accu_method == 'kahan_sum':
        aa = accupy.kahan_sum(a)
    else:
        raise ValueError(accu_method)

    return aa


def x_dot(a1, a2):
    """
    A much slower, but more accurate dot product
    (see https://github.com/nschloe/accupy)
    """
    a1 = a1.astype(float)
    a2 = a2.astype(float)
    return accupy.fdot(a1, a2)


def monkey_patch():
    numpy.x_sum = x_sum
    numpy.x_dot = x_dot



from tqdm import tqdm
from joblib import Parallel, delayed


def x_on_iter(llist, func, n_jobs=-1):
    all_items = Parallel(n_jobs=n_jobs)(delayed(func)(i) for i in tqdm(llist, total=len(llist)))
    return all_items


def x_on_keys(ddict, keys, func, n_jobs=-1):
    all_keys = list(keys)
    all_vals = Parallel(n_jobs=n_jobs)(delayed(func)(ddict[k]) for k in tqdm(all_keys, total=len(all_keys)))
    new_dict = dict(zip(all_keys, all_vals))
    return new_dict
